/*
 * @(#)WelltrisN.c
 *
 * Copyright 1993 - 2008  David A. Bagley, bagleyd@tux.org
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995  Q. Alex Zhao, azhao@cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Next methods file for Welltris */

#include "file.h"
#include "rngs.h"
#include "WelltrisNP.h"

#ifdef WINVER
#ifndef LOGPATH
#define LOGPATH "/usr/tmp"
#endif
#ifndef INIFILE
#define INIFILE "wwelltris.ini"
#endif

#define SECTION "setup"

#else

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif

static void resizeNextGame(NextWelltrisWidget w);
static void destroyNextGame(Widget old);
#if 0
static void sizeNextGame(NextWelltrisWidget w);
#endif
static Boolean setValuesNextGame(Widget current, Widget request, Widget renew);
static void quitNextGame(NextWelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void initializeNextGame(Widget request, Widget renew);
static void exposeNextGame(Widget renew, XEvent *event, Region region);
static void hideNextGame(NextWelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void enterNextGame(NextWelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void leaveNextGame(NextWelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void showNextNextGame(NextWelltrisWidget w,
	XEvent *event, char **args, int nArgs);

static char defaultTranslationsNextGame[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <KeyPress>n: ShowNext()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

static XtActionsRec actionsListNextGame[] =
{
	{(char *) "Quit", (XtActionProc) quitNextGame},
	{(char *) "Hide", (XtActionProc) hideNextGame},
	{(char *) "ShowNext", (XtActionProc) showNextNextGame},
	{(char *) "Enter", (XtActionProc) enterNextGame},
	{(char *) "Leave", (XtActionProc) leaveNextGame}
};

static XtResource resourcesNextGame[] =
{
	/* 18*6+4 */
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(NextWelltrisWidget, core.width),
	 XtRString, (caddr_t) "148"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(NextWelltrisWidget, core.height),
	 XtRString, (caddr_t) "148"},
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(NextWelltrisWidget, welltris.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(NextWelltrisWidget, welltris.background),
	 XtRString, (caddr_t) XtDefaultBackground},
	{XtNunits, XtCUnits, XtRInt, sizeof (int),
	 XtOffset(NextWelltrisWidget, welltris.units),
	 XtRString, (caddr_t) "4"},
	{XtNdiagonal, XtCDiagonal, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.diagonal),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmixed, XtCMixed, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.mixed),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbonus, XtCBonus, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, nextWelltris.newBonus),
	 XtRString, (caddr_t) "FALSE"},
	{XtNplainPiece, XtCPlainPiece, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.plainPiece),
	 XtRString, (caddr_t) "FALSE"},
	{XtNshowNext, XtCShowNext, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.showNext),
	 XtRString, (caddr_t) "TRUE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(NextWelltrisWidget, welltris.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextWelltrisWidget, welltris.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNrandomNumber, XtCRandomNumber, XtRInt, sizeof (int),
	 XtOffset(NextWelltrisWidget, nextWelltris.randomNumber),
	 XtRString, (caddr_t) "-1"},
	{XtNrandomRotation, XtCRandomRotation, XtRInt, sizeof (int),
	 XtOffset(NextWelltrisWidget, nextWelltris.randomRotation),
	 XtRString, (caddr_t) "-1"},
	{XtNrandomReflection, XtCRandomReflection, XtRInt, sizeof (int),
	 XtOffset(NextWelltrisWidget, nextWelltris.randomReflection),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(NextWelltrisWidget, welltris.select),
	 XtRCallback, (caddr_t) NULL}
};

NextWelltrisClassRec nextWelltrisClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "NextWelltris",	/* class name */
		sizeof (NextWelltrisRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) initializeNextGame,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListNextGame,	/* actions */
		XtNumber(actionsListNextGame),	/* num actions */
		resourcesNextGame,	/* resources */
		XtNumber(resourcesNextGame),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) destroyNextGame,	/* destroy */
		(XtWidgetProc) resizeNextGame,	/* resize */
		(XtExposeProc) exposeNextGame,	/* expose */
		(XtSetValuesFunc) setValuesNextGame,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsNextGame,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass nextWelltrisWidgetClass = (WidgetClass) & nextWelltrisClassRec;

static void
setNextGame(NextWelltrisWidget w, int reason)
{
	welltrisCallbackStruct cb;

	cb.reason = reason;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}
#endif

#include "image.h"

static Pixmap dr = 0; /* dummy for future double buffering */

static void
checkThings(NextWelltrisWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->welltris.units < MIN_UNITS) {
		w->welltris.units = MIN_UNITS;
		w->welltris.next.units = w->welltris.units;
		intCat(&buf1, "Number of units out of bounds, use at least ",
			MIN_UNITS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->welltris.units);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (w->welltris.units > MAX_UNITS) {
		w->welltris.units = MAX_UNITS;
		w->welltris.next.units = w->welltris.units;
		intCat(&buf1, "Number of units out of bounds, use at most ",
			MAX_UNITS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->welltris.units);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
}

static void
draw3DFrame(const NextWelltrisWidget w, Boolean inside,
	int x, int y, int width, int height)
{
	GC gc;

	if (inside)
		gc = w->welltris.inverseGC[3];
	else
		gc = w->welltris.inverseGC[1];
	FILLRECTANGLE(w, dr, gc,
		x, y, width, 1);
	FILLRECTANGLE(w, dr, gc,
		x, y + 1, 1, height - 1);
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + 1, width - 2, 1);
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + 2, 1, height - 3);
	if (inside)
		gc = w->welltris.inverseGC[0];
	else
		gc = w->welltris.inverseGC[1];
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + height - 1, width - 1, 1);
	FILLRECTANGLE(w, dr, gc,
		x + width - 1, y + 1, 1, height - 2);
	FILLRECTANGLE(w, dr, gc,
		x + 2, y + height - 2, width - 3, 1);
	FILLRECTANGLE(w, dr, gc,
		x + width - 2, y + 2, 1, height - 4);
}

static void
eraseFrame(NextWelltrisWidget w)
{
	FILLRECTANGLE(w, dr, (w->welltris.showNext) ?
		w->welltris.inverseGC[2] : w->welltris.inverseGC[1],
		w->welltris.offset.x, w->welltris.offset.y,
		w->welltris.viewLength, w->welltris.viewLength);
}

static void
drawFrame(NextWelltrisWidget w)
{
	int offsetX = w->welltris.offset.x - 2;
	int offsetY = w->welltris.offset.y - 2;
	int lengthX = w->welltris.offset.x + w->welltris.viewLength + 2;
	int lengthY = w->welltris.offset.y + w->welltris.viewLength + 2;

	if (w->welltris.showNext) {
		FILLRECTANGLE(w, dr, w->welltris.inverseGC[4],
			0, 0, w->core.width, offsetY);
		FILLRECTANGLE(w, dr, w->welltris.inverseGC[4],
			0, offsetY, offsetX, lengthY - offsetY);
		FILLRECTANGLE(w, dr, w->welltris.inverseGC[4],
			lengthX, offsetY,
			w->core.width - lengthX, lengthY - offsetY);
		FILLRECTANGLE(w, dr, w->welltris.inverseGC[4],
			0, lengthY, w->core.width, offsetY + 1);
		draw3DFrame(w, w->welltris.showNext,
			offsetX, offsetY,
			lengthX - offsetX, lengthY - offsetY);
	} else {
		FILLRECTANGLE(w, dr, w->welltris.inverseGC[1],
			w->welltris.offset.x, w->welltris.offset.y,
			w->welltris.viewLength, w->welltris.viewLength);
	}
}

static void
drawUnit(NextWelltrisWidget w, int pmid, int cid, int i, int j)
{
#ifdef IMAGE
	Boolean image = True;
#ifdef WINVER
	HBITMAP hOldBitmap, monoOldBitmap, hBitmap;
	HDC monoDC;
#endif

	if (images[TRUE][pmid % BITMAPS] == None) {
		if (images[TRUE][BITMAPS - 1] != None)
			(void) printf("Bitmap unknown %d (Next).\n", pmid);
		image = False;
	}
#ifdef WINVER
	if (image && (monoDC = CreateCompatibleDC(0)) == NULL) {
		(void) printf("Can not create mono DC (Next).\n");
		image = False;
	}
	if (image && (w->core.memDC =
			CreateCompatibleDC(w->core.hDC)) == NULL) {
		(void) printf("Can not create compatible DC (Next).\n");
		image = False;
	}
	if (image && (hBitmap = CreateCompatibleBitmap(w->core.hDC,
			w->welltris.boxSize, w->welltris.boxSize)) == NULL) {
		(void) printf("Can not create compatible bitmap (Next).\n");
		image = False;
	}
#endif
	if (image) {
#ifdef WINVER
		SetTextColor(w->core.memDC, w->welltris.inverseGC[2]);
		SetBkColor(w->core.memDC, w->welltris.thingGC[cid][2]);
		FILLRECTANGLE(w, hBitmap, w->welltris.thingGC[cid][2],
			0, 0,
			w->welltris.boxSize, w->welltris.boxSize);
		monoOldBitmap = (HBITMAP) SelectObject(monoDC,
			images[TRUE][pmid % BITMAPS]);
		hOldBitmap = SelectObject(w->core.memDC, hBitmap);
		BitBlt(w->core.memDC,
			0, 0, w->welltris.boxSize, w->welltris.boxSize,
			monoDC, 0, 0,
			SRCCOPY);
		(void) SelectObject(monoDC, monoOldBitmap);
		DeleteDC(monoDC);
		BitBlt(w->core.hDC,
			w->welltris.boxSize * i + w->welltris.offset.x,
			w->welltris.boxSize * j + w->welltris.offset.y,
			w->welltris.boxSize, w->welltris.boxSize,
			w->core.memDC, 0, 0,
			SRCCOPY);
		(void) SelectObject(w->core.memDC, hOldBitmap);
		DeleteObject(hBitmap);
		DeleteDC(w->core.memDC);
#else
		(void) XPutImage(XtDisplay(w), XtWindow(w),
			w->welltris.thingGC[cid][2],
			images[TRUE][pmid % BITMAPS],
			0, 0,
			w->welltris.boxSize * i + w->welltris.offset.x,
			w->welltris.boxSize * j + w->welltris.offset.y,
			w->welltris.boxSize, w->welltris.boxSize);
#endif
	} else
#endif
	{
		FILLRECTANGLE(w, dr, w->welltris.thingGC[cid][2],
			w->welltris.boxSize * i + w->welltris.offset.x,
			w->welltris.boxSize * j + w->welltris.offset.y,
			w->welltris.boxSize, w->welltris.boxSize);
	}
}

static void
clearThing(NextWelltrisWidget w)
{
	FILLRECTANGLE(w, dr, w->welltris.inverseGC[2],
		w->welltris.offset.x, w->welltris.offset.y,
		w->welltris.viewLength, w->welltris.viewLength);
}

static void
drawThing(NextWelltrisWidget w)
{
	int i, j;

	if (w->welltris.nextThing.polyNumber == IGNORE_RAND)
		return;
	clearThing(w);
	drawFrame(w);
	for (i = 0; i < w->welltris.nextThing.size; i++)
		for (j = 0; j < w->welltris.nextThing.size; j++)
			if (tris[w->welltris.nextThing.units].polyomino
					[w->welltris.nextThing.polyNumber].shape[j][i] != 0) {
				drawUnit(w, polyPixmap(i, j,
					(tris[w->welltris.nextThing.units].polyomino
					[w->welltris.nextThing.polyNumber].shape),
					w->welltris.nextThing.size),
					w->welltris.nextThing.colorNumber,
					w->welltris.nextThing.xpos + i,
					w->welltris.nextThing.ypos + j);
			}
}

static void
redoNextNext(NextWelltrisWidget w)
{
	redoNext((WelltrisWidget) w);
	w->welltris.nextThing.xpos =
		(MAX_UNITS - w->welltris.nextThing.size) / 2;
	w->welltris.nextThing.ypos =
		(MAX_UNITS - w->welltris.nextThing.size) / 2;
#ifdef DEBUG
	(void) printf("Size: %d, Poly: %d, Color %d\n",
		w->welltris.nextThing.size, w->welltris.nextThing.polyNumber,
		w->welltris.nextThing.colorNumber);
#endif
}

#if 0
static void
resetThings(NextWelltrisWidget w)
{
}
#endif

static void
resizeThings(NextWelltrisWidget w)
{
#ifdef IMAGE
#ifndef WINVER
	Display *display = XtDisplay(w);
	Window window = XtWindow(w);
	XWindowAttributes xgwa;

	(void) XGetWindowAttributes(display, window, &xgwa);
#endif
	free_images(TRUE);
	if (w->welltris.boxSize > 4)
		(void) create_images(
#ifndef WINVER
			display, xgwa.visual,
#endif
			(w->welltris.plainPiece) ? 0 : 1, TRUE,
			w->welltris.boxSize);
#endif
}

#ifdef WINVER
static void
setValuesNextGame(NextWelltrisWidget w)
{
	struct tagColor {
		int red, green, blue;
	} color;
	char szBuf[80], buf[20];
	int thing;

	w->welltris.units = GetPrivateProfileInt(SECTION, "units",
		DEFAULT_UNITS, INIFILE);
	w->welltris.diagonal = (BOOL) GetPrivateProfileInt(SECTION, "diagonal",
		DEFAULT_DIAGONAL, INIFILE);
	w->welltris.mixed = (BOOL) GetPrivateProfileInt(SECTION, "mixed",
		DEFAULT_MIXED, INIFILE);
	w->welltris.plainPiece = (BOOL) GetPrivateProfileInt(SECTION,
		"plainPiece", FALSE, INIFILE);
	w->welltris.showNext = (BOOL) GetPrivateProfileInt(SECTION, "showNext",
		FALSE, INIFILE);
	w->welltris.mono = (BOOL) GetPrivateProfileInt(SECTION, "mono",
		DEFAULT_MONO, INIFILE);
	w->welltris.reverse = (BOOL) GetPrivateProfileInt(SECTION,
		"reverseVideo", DEFAULT_REVERSE, INIFILE);
	/* #AEB2C3 */
	(void) GetPrivateProfileString(SECTION, "background", "255 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->welltris.inverseGC[1] = RGB(color.red, color.green, color.blue);
	w->welltris.inverseGC[0] = RGB(brighter(color.red),
		brighter(color.green), brighter(color.blue));
	w->welltris.inverseGC[2] = RGB(darker(color.red),
		darker(color.green), darker(color.blue));
	w->welltris.inverseGC[3] = RGB(darker(darker(color.red)),
		darker(darker(color.green)), darker(darker(color.blue)));
	w->welltris.inverseGC[4] = RGB(darker(darker(darker(color.red))),
		darker(darker(darker(color.green))),
		darker(darker(darker(color.blue))));
	for (thing = 0; thing < MAX_START_POLYOMINOES; thing++) {
		(void) sprintf(buf, "thingColor%d", thing);
		(void) GetPrivateProfileString(SECTION, buf,
			polyominoColorString[thing],
			szBuf, sizeof (szBuf), INIFILE);
		(void) sscanf(szBuf, "%d %d %d",
			&(color.red), &(color.green), &(color.blue));
		w->welltris.thingGC[thing][2] =
			RGB(color.red, color.green, color.blue);
	}
}

void
destroyNextGame(HBRUSH brush)
{
	(void) DeleteObject(brush);
	/*PostQuitMessage(0);*/
}

#else

static void
getColor(NextWelltrisWidget w, int thing)
{
	XGCValues values;
	XtGCMask valueMask;
	XColor colorCell, rgb;
	Boolean gotColor = False;

	valueMask = GCForeground | GCBackground;
	if (w->welltris.reverse) {
		values.background = w->welltris.foreground;
	} else {
		values.background = darker((WelltrisWidget) w, w->welltris.background);
	}
	if (!w->welltris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyominoColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->welltris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyominoColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyomino ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->welltris.reverse) {
			values.foreground = w->welltris.background;
		} else {
			values.foreground = w->welltris.foreground;
		}
	}
	if (w->welltris.thingGC[thing][2])
		XtReleaseGC((Widget) w, w->welltris.thingGC[thing][2]);
	w->welltris.thingGC[thing][2] = XtGetGC((Widget) w, valueMask, &values);
}

static void
setAllColors(NextWelltrisWidget w)
{
	XGCValues values;
	XtGCMask valueMask;
	int polyomino;

	valueMask = GCForeground | GCBackground;
	if (w->welltris.reverse) {
		values.background = w->welltris.background;
		values.foreground = w->welltris.foreground;
	} else {
		values.foreground = w->welltris.background;
		values.background = w->welltris.foreground;
	}
	if (w->welltris.inverseGC[1])
		XtReleaseGC((Widget) w, w->welltris.inverseGC[1]);
	w->welltris.inverseGC[1] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->welltris.mono) {
		values.foreground = brighter((WelltrisWidget) w, (w->welltris.reverse) ?
			w->welltris.foreground : w->welltris.background);
	}
	if (w->welltris.inverseGC[0])
		XtReleaseGC((Widget) w, w->welltris.inverseGC[0]);
	w->welltris.inverseGC[0] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->welltris.mono) {
		values.foreground = darker((WelltrisWidget) w, (w->welltris.reverse) ?
			w->welltris.foreground : w->welltris.background);
	}
	if (w->welltris.inverseGC[2])
		XtReleaseGC((Widget) w, w->welltris.inverseGC[2]);
	w->welltris.inverseGC[2] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->welltris.mono) {
		values.foreground = darker((WelltrisWidget) w, darker((WelltrisWidget) w, (w->welltris.reverse) ?
			w->welltris.foreground : w->welltris.background));
	}
	if (w->welltris.inverseGC[3])
		XtReleaseGC((Widget) w, w->welltris.inverseGC[3]);
	w->welltris.inverseGC[3] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->welltris.mono) {
		values.foreground = darker((WelltrisWidget) w, darker((WelltrisWidget) w, darker((WelltrisWidget) w,
			(w->welltris.reverse) ? w->welltris.foreground :
			w->welltris.background)));
	}
	if (w->welltris.inverseGC[4])
		XtReleaseGC((Widget) w, w->welltris.inverseGC[4]);
	w->welltris.inverseGC[4] = XtGetGC((Widget) w, valueMask, &values);
	for (polyomino = 0; polyomino < MAX_START_POLYOMINOES; polyomino++)
		getColor(w, polyomino);
}

static Boolean
setValuesNextGame(Widget current, Widget request, Widget renew)
{
	NextWelltrisWidget c = (NextWelltrisWidget) current;
	NextWelltrisWidget w = (NextWelltrisWidget) renew;
	Boolean redraw = False, setColors = False;

	checkThings(w);
	if (w->welltris.showNext != c->welltris.showNext) {
		redraw = True;
	}
	if (w->welltris.reverse != c->welltris.reverse ||
			w->welltris.mono != c->welltris.mono) {
		setAllColors(w);
		redraw = True;
	} else if (w->welltris.background != c->welltris.background ||
			w->welltris.foreground != c->welltris.foreground ||
			setColors) {
		setAllColors(w);
		redraw = True;
	} else if (w->welltris.plainPiece != c->welltris.plainPiece) {
		redraw = True;
	}
#if 0
	if (w->welltris.units != c->welltris.units) {
		w->welltris.next.units = w->welltris.units;
		resetThings(w);
	} else if (w->welltris.diagonal != c->welltris.diagonal) {
		w->welltris.next.diagonal = w->welltris.diagonal;
		resetThings(w);
	} else if (w->welltris.mixed != c->welltris.mixed) {
		w->welltris.next.mixed = w->welltris.mixed;
		resetThings(w);
	}
#endif
	if (w->welltris.menu != -1) {
		switch (w->welltris.menu) {
		case 3:
			w->welltris.menu = -1;
			clearThing(w);
			redraw = True;
			break;
		default:
			w->welltris.menu = -1;
			break;
		}
	}
	if (w->nextWelltris.randomNumber == CLEARNEXT_RAND) {
		if (w->welltris.showNext) {
			drawThing(w);
		} else {
			clearThing(w);
		}
		w->nextWelltris.randomNumber = IGNORE_RAND;
		redraw = True;
	} else if (w->nextWelltris.randomNumber != IGNORE_RAND) {
		w->welltris.nextThing.randomNumber = w->nextWelltris.randomNumber;
		w->welltris.nextThing.randomRotation = w->nextWelltris.randomRotation;
		w->welltris.nextThing.randomReflection = w->nextWelltris.randomReflection;
		w->welltris.nextThing.bonus = w->nextWelltris.newBonus;
		w->welltris.next.units = w->welltris.units;
		w->welltris.next.diagonal = w->welltris.diagonal;
		w->welltris.next.mixed = w->welltris.mixed;
		redoNextNext(w);
		if (w->welltris.showNext) {
			drawThing(w);
		}
		w->nextWelltris.randomNumber = IGNORE_RAND;
		redraw = True;
	}
	return (redraw);
}

static void
quitNextGame(NextWelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	Display *display = XtDisplay(w);

	XtCloseDisplay(display);
	exit(0);
}

static void
destroyNextGame(Widget old)
{
	NextWelltrisWidget w = (NextWelltrisWidget) old;
	int thing, i;

	for (thing = 0; thing < MAX_START_POLYOMINOES; thing++)
		XtReleaseGC(old, w->welltris.thingGC[thing][2]);
	for (i = 0; i < BG_SHADES; i++)
		XtReleaseGC(old, w->welltris.inverseGC[i]);
	XtRemoveCallbacks(old, XtNselectCallback, w->welltris.select);
}
#endif

#ifndef WINVER
static
#endif
void
resizeNextGame(NextWelltrisWidget w)
{
#ifdef WINVER
	RECT rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom;
#endif
	w->welltris.boxSize = (MIN(w->core.height, w->core.width) - 4) /
		MAX_UNITS;
	w->welltris.viewLength = w->welltris.boxSize * MAX_UNITS;
	w->welltris.offset.x = ((int) w->core.width + 1 -
		w->welltris.viewLength) / 2;
	w->welltris.offset.y = ((int) w->core.height + 1 -
		w->welltris.viewLength) / 2;
}

#if 0
#ifndef WINVER
static
#endif
void
sizeNextGame(NextWelltrisWidget w)
{
	resetThings(w);
	resizeNextGame(w);
}
#endif

#ifndef WINVER
static
#endif
void
initializeNextGame(
#ifdef WINVER
NextWelltrisWidget w, HBRUSH brush
#else
Widget request, Widget renew
#endif
)
{
	int diagonal, mixed, turnStyle;
#ifdef IMAGE
	int bitmap;
#endif
#ifndef WINVER
	int thing, i;
	NextWelltrisWidget w = (NextWelltrisWidget) renew;
#endif

#ifdef WINVER
	setValuesNextGame(w);
#else
	w->welltris.mono = (DefaultDepthOfScreen(XtScreen(w)) < 2 ||
		w->welltris.mono);
	for (thing = 0; thing < MAX_START_POLYOMINOES; thing++)
		w->welltris.thingGC[thing][2] = NULL;
	for (i = 0; i < BG_SHADES; i++)
		w->welltris.inverseGC[i] = NULL;
#endif
#ifdef IMAGE
	for (bitmap = 0; bitmap < BITMAPS; bitmap++)
		images[TRUE][bitmap] = None;
#endif

	w->welltris.quit = False;
	w->welltris.focus = False;
	w->welltris.next.units = w->welltris.units;
	w->welltris.next.diagonal = w->welltris.diagonal;
	w->welltris.next.mixed = w->welltris.mixed;
	w->welltris.nextThing.polyNumber = IGNORE_RAND;
	checkThings(w);
#if 0
	sizeNextGame(w);
#else
	resizeNextGame(w);
#endif
#ifdef WINVER
	brush = CreateSolidBrush(w->welltris.inverseGC[2]);
	SETBACK(w->core.hWnd, brush);
	(void) SRAND(time(NULL));
#else
	(void) SRAND(getpid());
	setAllColors(w);
#endif
	diagonal = (w->welltris.next.diagonal) ? 1 : 0;
	mixed = (w->welltris.next.mixed) ? 1 : 0;
	readThings();
	turnStyle = tris[w->welltris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	/* use intCat */
	if (turnStyle < 0 || turnStyle >= MAX_TURNSTYLE_TYPES) {
		(void) fprintf(stderr, "NextWelltris: corrupted polyomino file\n");
		(void) fprintf(stderr, "\tturnStyle = %d.\n", turnStyle);
		exit(1);
	}
	w->welltris.numberPolyominoes = tris[w->welltris.next.units -
		MIN_UNITS].mode[diagonal].number[mixed][turnStyle];
	if (w->welltris.numberPolyominoes <= 0 ||
			w->welltris.numberPolyominoes > MAX_START_POLYOMINOES) {
		(void) fprintf(stderr, "NextWelltris: corrupted polyomino file\n");
		(void) fprintf(stderr, "\tnumberPolyominoes = %d.\n",
			w->welltris.numberPolyominoes);
		exit(1);
	}
}

#ifndef WINVER
static
#endif
void
exposeNextGame(
#ifdef WINVER
NextWelltrisWidget w
#else
Widget renew, XEvent *event, Region region
#endif
)
{
#ifndef WINVER
	NextWelltrisWidget w = (NextWelltrisWidget) renew;

	if (!w->core.visible)
		return;
#endif
	eraseFrame(w);
	resizeThings(w);
	drawFrame(w);
	if (w->welltris.showNext) {
		drawThing(w);
	}
}

#ifndef WINVER
static
#endif
void
hideNextGame(NextWelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setNextGame(w, ACTION_HIDE);
}

#ifdef WINVER
void
valueNextGame(NextWelltrisWidget w,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean diagonal, Boolean mixed, Boolean newBonus)
{
	w->welltris.nextThing.randomNumber = randomNumber;
	w->welltris.nextThing.randomRotation = randomRotation;
	w->welltris.nextThing.randomReflection = randomReflection;
	w->welltris.next.units = w->welltris.units = units;
	w->welltris.next.diagonal = w->welltris.diagonal = diagonal;
	w->welltris.next.mixed = w->welltris.mixed = mixed;
	w->welltris.nextThing.bonus = newBonus;
	redoNextNext(w);
	if (w->welltris.showNext) {
		drawThing(w);

	}
}
#endif

#ifndef WINVER
static
#endif
void
enterNextGame(NextWelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->welltris.focus = True;
	drawFrame(w);
}

#ifndef WINVER
static
#endif
void
leaveNextGame(NextWelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->welltris.focus = False;
	drawFrame(w);
}

#ifndef WINVER
static
#endif
void
showNextNextGame(NextWelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->welltris.paused) {
		w->welltris.showNext = !w->welltris.showNext;
		if (w->welltris.showNext) {
			drawThing(w);
		} else {
			clearThing(w);
			drawFrame(w);
		}
		FLUSH(w);
	}
}
